/*
 * biod.c	Central I/O functions for nfsbench
 *
 * A `biod' is an abstraction of a process handling a single RPC operation.
 * A socket may have one or more biods acting on it. For each socket, the
 * numbers of idle and busy biods associated with that socket are maintained
 * in biod_idle[] and biod_busy[] arrays. I found that for a small number
 * of sockets, searching an array is faster than dealing with linked lists
 * and stuff.
 *
 * The overall approach is to send as many RPC requests as possible, and
 * then wait for replies. When there's a read condition on a socket, all
 * replies are read and processed before another bunch of requests is
 * generated. It seems to be faster to do it this way (i.e. send, send, ...,
 * select, recv, recv, ..., send, send, ...) rather than select, recv, send,
 * select, recv, send, etc.
 *
 * Requests are kept in a doubly linked list, ordered by their estimated time
 * of arrival. If the reply doesn't arrive within that time, retransmission
 * kicks in.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/time.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

#include "nfsbench.h"
#undef	VERBOSE

#ifndef NR_OPEN
# ifdef __hpux
#  define NR_OPEN	NOFILE
# else
#  error "Don't know how to define NR_OPEN"
# endif
#endif

static int		biod_fds[MAXBIODS];
static fd_set		biod_fdset;
static int		biod_avail;
static int		biod_idle[NR_OPEN];
static int		biod_busy[NR_OPEN];
static int		biod_draining = 0;
static int		biod_synchronous = 0;

static void		do_output(void);
static void		do_rexmit(void);
static void		do_input(int fd);
static void		biod_init(void);


/*
 * Output as many requests as possible.
 * We don't do any sort of congestion control yet.
 */
static void
do_output(void)
{
	struct nfsreq	*req;
	unsigned long	now;

	while (biod_avail) {
		static int	lastwfd = 0;
		int		err, fd;

		while (!(req = req_queued)) {
			if (biod_draining)
				return;
			if (!script_nextaction()) {
				/* script is done */
				return;
			}
		}

		while (!biod_idle[(fd = biod_fds[lastwfd])]) {
			if (++lastwfd >= nrsocks)
				lastwfd = 0;
		}

		/* Check input - this is a hack */
		do_input(fd);

#ifdef VERBOSE
		printf("sending %08lx on sock %d\n", req->rq_xid, fd);
		if (req->rq_next)
			printf("next is %08lx\n", req->rq_next->rq_xid);
#if VERBOSE > 2
		{
		   u32 *m = req->rq_msg;
		   printf(
		   "000 %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n"
		   "020 %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n"
		   "040 %08lx %08lx %08lx %08lx %08lx %08lx %08lx %08lx\n",
		   m[0],  m[1],  m[2],  m[3],  m[4],  m[5],  m[6],  m[7],
		   m[8],  m[9],  m[10], m[11], m[12], m[13], m[14], m[15],
		   m[16], m[17], m[18], m[19], m[20], m[21], m[22], m[23]);
		}
#endif
#endif
		now = jiffies();
		err = sendto(fd, req->rq_msg, req->rq_msglen, 0,
				(struct sockaddr *) req->rq_addr,
				sizeof(*(req->rq_addr)));
		if (err < 0) {
			if (errno == EWOULDBLOCK || errno == ENOBUFS
			 || errno == EAGAIN)
				break;
			perror("sendto");
			exit(1);
		}
		/* Move biod to busy queue */
		biod_idle[fd]--;
		biod_busy[fd]++;
		biod_avail--;

		/* set timeout and move request to pending queue */
		req->rq_sock = fd;
		if (!req->rq_qtime)
			req->rq_qtime = now;
		req->rq_expires = now + req->rq_timeout;

		if (!(req_queued = req->rq_next))
			req_queued_tail = NULL;
		wait_request(req);

		stats.byte_sent += req->rq_msglen;
		stats.pkt_sent++;
	}

#if 0
	/* Generate some more requests while packets are in flight */
	if (req_free && !biod_draining)
		script_nextaction();
#endif
}

static void
do_rexmit(void)
{
	struct nfsreq	*req;
	unsigned long	now;

	now = jiffies();
	while ((req = req_pending) != NULL && req->rq_expires < now) {
		if (!(req_pending = req->rq_next))
			req_pending_tail = NULL;
		if (++(req->rq_rexmits) > MAXREXMITS) {
#ifdef VERBOSE
			printf("request %08lx timed out\n",
				(unsigned long) req->rq_xid);
#endif
			finish_request(req, NULL, 0);
			stats.pkt_lost++;
			biod_busy[req->rq_sock]--;
			biod_idle[req->rq_sock]++;
			biod_avail++;
		} else {
			req->rq_timeout <<= 1;
			req->rq_expires = now + req->rq_timeout;
			sendto(req->rq_sock, req->rq_msg, req->rq_msglen, 0,
				(struct sockaddr *) req->rq_addr,
				sizeof(*(req->rq_addr)));
			wait_request(req);
			stats.pkt_rxmt++;
#ifdef VERBOSE
			printf("request %08lx retransmitted\n",
				(unsigned long) req->rq_xid);
#endif
		}
	}
}

static void
do_input(int fd)
{
	struct nfsreq	*req, *prv;
	u32		buffer[MAXPKTSIZE/4], xid;
	unsigned long	now;
	int		len;

	while (biod_busy[fd]) {
		if ((len = recv(fd, buffer, MAXPKTSIZE, 0)) < 0) {
			if (errno == EWOULDBLOCK || errno == EAGAIN)
				return;
			fprintf(stderr, "socket %d: ", fd);
			perror("recv");
			exit(2);
		}

		now = jiffies();
		xid = buffer[0];
		prv = NULL;
		for (req = req_pending; req; prv = req, req = req->rq_next) {
			if (req->rq_xid == xid && req->rq_sock == fd)
				break;
		}

		if (!req) {
			/* Server reply was duplicated or link was slow so
			 * we retransmitted too fast */
			/*
			printf("unknown reply sock %d xid %08lx\n",
				fd, (unsigned long) buffer[0]);
			 */
			return;
		}
		if (!prv) {
			req_pending = req->rq_next;
		} else {
			prv->rq_next = req->rq_next;
		}
		if (!req->rq_next)
			req_pending_tail = prv;

		biod_busy[fd]--;
		biod_idle[fd]++;
		biod_avail++;
#ifdef VERBOSE
		printf("finish_request = %08lx %d jiffies\n",
				req->rq_xid, jiffies() - req->rq_qtime);
#endif
		req->rq_rtt = now - req->rq_qtime;
		finish_request(req, buffer, len);

		stats.byte_recv += len;
		stats.pkt_recv++;
		stats.rtt += req->rq_rtt;
		if (req->rq_rtt > stats.max_rtt)
			stats.max_rtt = req->rq_rtt;
		if (req->rq_rtt < stats.min_rtt)
			stats.min_rtt = req->rq_rtt;
	}
}

void
biod(void)
{
	static int	inited = 0;
	struct timeval	tv;
	fd_set		fdset;
	unsigned long	now, timeo;
	int		i;

	if (!inited) {
		biod_init();
		inited = 1;
	}

	while (1) {
		if (biod_avail)
			do_output();

		if (!req_pending) {
			if (!biod_draining || biod_synchronous) {
				/* Test script apparently done */
				return;
			}
			/* Output drained */
			if (!req_queued)
				biod_draining = 0;
			continue;
		}

#if 0
		for (i = 0; i < nrsocks; i++) {
			do_input(biod_fds[i]);
		}
		if (!req_pending)
			continue;
#endif

		if ((now = jiffies()) >= req_pending->rq_expires) {
			do_rexmit();
			continue;
		}

		/*
		 * This strange loop is a hack for profiling... Did you
		 * know you also have to profile benchmarks?
		 */
again:
		timeo = req_pending->rq_expires - now;
		tv.tv_sec = timeo / HZ;
		tv.tv_usec = (timeo * 1000000 / HZ) % 1000000;
		fdset = biod_fdset;
		/* printf("timeo %d.%d\n", tv.tv_sec, tv.tv_usec); */
		if (select(FD_SETSIZE, &fdset, NULL, NULL, &tv) < 0) {
			if (errno == EINTR) {
				time(&now);
				goto again;
			}
			perror("select");
			exit(2);
		}

		for (i = 0; i < nrsocks; i++) {
			int	fd = biod_fds[i];

			if (FD_ISSET(fd, &fdset))
				do_input(fd);
		}

		/* Do any necessary retransmits */
		if (req_pending && (now = jiffies()) >= req_pending->rq_expires)
			do_rexmit();
	}
}

void
biod_drain_on(void)
{
	biod_draining = 1;
}

void
biod_drain(void)
{
	if (req_pending || req_queued) {
		biod_draining = 1;
		biod();
	}
}

void
biod_flush(void)
{
	biod_synchronous = 1;
	biod_drain();
	biod_synchronous = 0;
}

static void
biod_init(void)
{
	unsigned long	sz = 64 * 1024;
	int		fd = -1, i, j, spb;

	spb = (nrsocks + nrbiods - 1) / nrbiods;
	FD_ZERO(&biod_fdset);
	for (i = 0, j = 0; i < nrbiods; i++) {
		if (fd < 0 || ((i + 1) % spb) != 0) {
			fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
			if (fd < 0) {
				perror("socket");
				exit(2);
			}
			if (fd >= NR_OPEN) {
				fprintf(stderr, "Huh? fd %d > NR_OPEN\n", fd);
				exit(2);
			}
			setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sz, sizeof(sz));
			setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sz, sizeof(sz));
			fcntl(fd, F_SETFL, O_NONBLOCK);
			FD_SET(fd, &biod_fdset);
			biod_fds[j++] = fd;
			biod_idle[fd] = 0;
			biod_busy[fd] = 0;
		}
		biod_idle[fd]++;
		biod_avail++;
	}

	nrsocks = j;
}

