/*
 * dir.c	directory operations for nfsbench
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <string.h>
#include "nfsbench.h"
#include "fs/nfs.h"

static void
nfs_create_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_create_res(buffer, &vn->n_fh, &vn->n_fattr);
	if (nfs_check("create", vn->n_path)) {
		vn->n_flags |= N_VALIDFH | N_VALIDFATTR;
		vn->n_type = vn->n_fattr.type;
	}
	nfs_notify(vn);
	nfs_vput(vn);
}

struct nfsnode *
nfs_create(struct nfsnode *vn, char *name, struct nfs_sattr *sattr)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	if (!(ch = nfs_lookup_cache(vn, name)))
		return NULL;

	if (!(req = generate_request())) {
		nfs_vput(ch);
		return NULL;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_create(req->rq_msg, &vn->n_fh, name, sattr);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_create_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	return ch;
}

static void
nfs_mkdir_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_mkdir_res(buffer, &vn->n_fh, &vn->n_fattr);
	if (nfs_check("mkdir", vn->n_path)) {
		vn->n_flags |= N_VALIDFH | N_VALIDFATTR;
		vn->n_type = vn->n_fattr.type;
	}
	nfs_notify(vn);
	nfs_vput(vn);
}

struct nfsnode *
nfs_mkdir(struct nfsnode *vn, char *name, struct nfs_sattr *sattr)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	if (!(ch = nfs_lookup_cache(vn, name)))
		return NULL;

	if (!(req = generate_request())) {
		nfs_vput(ch);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_mkdir(req->rq_msg, &vn->n_fh, name, sattr);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_mkdir_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	return ch;
}

static void
nfs_remove_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_remove_res(buffer);
	if (nfs_check("remove", vn->n_path)) {
		nfs_notify(vn);
		nfs_vremove(vn);
	} else {
		nfs_notify(vn);
		nfs_vput(vn);
	}
}

int
nfs_remove(struct nfsnode *vn, char *name)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	if (!(ch = nfs_lookup_cache(vn, name)))
		return 0;

	if (/* ch->n_count != 1 || */ !(ch->n_flags & N_VALIDFH)) {
		nfs_vput(ch);
		return 0;
	}

	if (!(req = generate_request())) {
		nfs_vput(ch);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_remove(req->rq_msg, &vn->n_fh, name);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_remove_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	ch->n_flags |= N_DELETED;
	return 1;
}

static void
nfs_rmdir_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_rmdir_res(buffer);
	if (nfs_check("rmdir", vn->n_path)) {
		nfs_notify(vn);
		nfs_vremove(vn);
	} else {
		nfs_notify(vn);
		nfs_vput(vn);
	}
}

int
nfs_rmdir(struct nfsnode *vn, char *name)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	if (!(ch = nfs_lookup_cache(vn, name)))
		return 0;

	if (/* ch->n_count != 1 || */ !(ch->n_flags & N_VALIDFH)) {
		script_error("nfs_rmdir invalid fh for %s\n", ch->n_path);
		nfs_vput(ch);
		return 0;
	}

	if (!(req = generate_request())) {
		nfs_vput(ch);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_rmdir(req->rq_msg, &vn->n_fh, name);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_rmdir_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	ch->n_flags |= N_DELETED;
	return 1;
}

static void
nfs_readlink_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;
	char		path[NFS_MAXPATHLEN+1];

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_readlink_res(buffer, path);
	if (nfs_check("readlink", vn->n_path)) {
		vn->n_symlink = xstrdup(path);
		nfs_notify(vn);
	}
	nfs_vput(vn);
}

int
nfs_readlink(struct nfsnode *vn)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request())) {
		nfs_vput(vn);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_readlink(req->rq_msg, &vn->n_fh);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_readlink_reply;
	req->rq_clntdata = vn;

	enqueue_request(req);

	return 1;
}

static void
nfs_symlink_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_symlink_res(buffer);
	if (nfs_check("symlink", vn->n_path)) {
		nfs_notify(vn);
		vn->n_type = NFLNK;
	}
	nfs_vput(vn);
}

struct nfsnode *
nfs_symlink(struct nfsnode *vn, char *name, char *path, struct nfs_sattr *sattr)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	/* printf("nfs_symlink(%s, %s)\n", vn->n_path, name); */
	if (!(ch = nfs_lookup_cache(vn, name)))
		return NULL;

	if (!(req = generate_request())) {
		nfs_vput(ch);
		return NULL;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_symlink(req->rq_msg, &vn->n_fh, name, path, sattr);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_symlink_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	return ch;
}

static void
nfs_rename_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_rename_res(buffer);
	if (nfs_check("rename", vn->n_path))
		nfs_notify(vn);
	nfs_vput(vn);
}

struct nfsnode *
nfs_rename(struct nfsnode *fvn, char *fname, struct nfsnode *tvn, char *tname)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	/*
	printf("nfs_rename(%s/%s, %s/%s)\n", fvn->n_path, fname,
						tvn->n_path, tname);
	 */
	if (!(ch = nfs_lookup_cache(tvn, tname))) {
		nfs_vput(fvn);
		return NULL;
	}

	if (!(req = generate_request())) {
		nfs_vremove(ch);
		nfs_vput(fvn);
		return NULL;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_rename(req->rq_msg, &fvn->n_fh, fname,
					  &tvn->n_fh, tname);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_rename_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	nfs_vput(fvn);
	return ch;
}

static void
nfs_link_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_link_res(buffer);
	if (nfs_check("link", vn->n_path))
		nfs_notify(vn);
	nfs_vput(vn);
}

struct nfsnode *
nfs_link(struct nfsnode *fvn, struct nfsnode *tvn, char *name)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	/* printf("nfs_link(%s, %s/%s)\n", fvn->n_path, tvn->n_path, name); */
	if (!(ch = nfs_lookup_cache(tvn, name))) {
		nfs_vput(fvn);
		return NULL;
	}

	if (!(req = generate_request())) {
		nfs_vremove(ch);
		nfs_vput(fvn);
		return NULL;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_link(req->rq_msg, &fvn->n_fh, &tvn->n_fh, name);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_link_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	nfs_vput(fvn);
	return ch;
}

struct nfsdir *
nfs_opendir(struct nfsnode *vn)
{
	struct nfsdir	*dir;

	if (vn->n_type != NFDIR) {
		script_error("nfs_opendir: %s is not a directory\n",
			vn->n_path);
		return NULL;
	}
	dir = (struct nfsdir *) xmalloc(sizeof(*dir));
	dir->d_vnode = vn;
	dir->d_cookie = 0;
	dir->d_eof = 0;

	return dir;
}

void
nfs_closedir(struct nfsdir *dir)
{
	nfs_vput(dir->d_vnode);
	xfree(dir);
}

static void
nfs_dirent_add(struct nfsdir *dir, char *name, u32 cookie)
{
	struct nfsnode	*ch;

	if (!name) {
		dir->d_eof = 1;
		return;
	}
	if (!strcmp(name, ".") || !strcmp(name, ".."))
		return;
	dir->d_vnode->n_count++;
	if (!(ch = nfs_lookup_cache(dir->d_vnode, name))) {
		script_error("nfs_dirent_add: can't add %s/%s to cache\n",
				dir->d_vnode->n_path, name);
		return;
	}
	/* printf("dirent_add %s\n", ch->n_path); */
	dir->d_cookie = cookie;
	nfs_vput(ch);
}

static void
nfs_readdir_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsdir	*dir;

	dir = (struct nfsdir *) req->rq_clntdata;
	xdr_nfs_readdir_res(buffer, (nfs_dirent_fn_t) nfs_dirent_add, dir);
	nfs_check("readdir", dir->d_vnode->n_path);
	dir->d_pend = 0;
	dir_notify(dir);
}

int
nfs_readdir(struct nfsdir *dir)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request()))
		return 0;

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_readdir(req->rq_msg, &dir->d_vnode->n_fh,
				dir->d_cookie, nfs_rsize);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_readdir_reply;
	req->rq_clntdata = dir;

	enqueue_request(req);

	dir->d_pend = 1;
	return 1;
}
