/*
 * file.c	NFS file functions
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <sys/param.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "nfsbench.h"
#include "fs/nfs.h"

struct nfsnode *
nfs_open(char *path, unsigned int mode)
{
	struct nfsnode	*vn;

	if (!(vn = nfs_getroot())) {
		script_error("couldn't get NFS root fh\n");
		return NULL;
	}
	if (!(vn = nfs_namei(vn, path))) {
		if (!(mode & O_CREAT)) {
			script_error("namei of %s failed\n", path);
			return NULL;
		}
		if ((vn = nfs_creat(path, mode)) == NULL) {
			script_error("couldn't create %s\n", path);
			return NULL;
		}
		return vn;
	}

	if ((mode & O_TRUNC) && !nfs_truncate(vn, 0)) {
		script_error("couldn't truncate %s\n", path);
		nfs_vput(vn);
		return NULL;
	}

	return vn;
}

struct nfsnode *
nfs_creat(char *cpath, unsigned int mode)
{
	struct nfsnode	*vn;
	struct nfs_sattr sattr;
	char		name[NFS_MAXNAMLEN+1];
	char		path[NFS_MAXPATHLEN];
	char		*sp, *cp;

	if (!(vn = nfs_getroot())) {
		script_error("couldn't get NFS root fh\n");
		return NULL;
	}
	strcpy(path, cpath);
	for (sp = path, cp = path; *sp; sp++) {
		if (*sp == '/')
			cp = sp+1;
	}
	strncpy(name, cp, NFS_MAXNAMLEN);
	name[NFS_MAXNAMLEN] = '\0';
	*cp = '\0';
	if (!(vn = nfs_namei(vn, path))) {
		script_error("namei of %s failed\n", path);
		return NULL;
	}
	if (!(vn->n_flags & N_VALIDFATTR)) {
		biod_sync(nfs_getattr(vn));
		if (!(vn->n_flags & N_VALIDFATTR)) {
			script_error("namei of \"%s\" failed\n", path);
			goto fail;
		}
	}

	if (vn->n_type != NFDIR) {
		script_error("\"%s\" is not a directory\n", path);
		goto fail;
	}

	nfs_clear_sattr(&sattr);
	if (mode & O_TRUNC)
		sattr.size = 0;
	sattr.mode = 0666;

	biod_sync(vn = nfs_create(vn, name, &sattr));
	if (!vn)
		return NULL;

	if (!(vn->n_flags & N_VALIDFH)) {
		script_error("create of %s/%s failed\n", path, name);
		goto fail;
	}

	return vn;

fail:
	if (vn)
		nfs_vput(vn);
	return NULL;
}

void
nfs_close(struct nfsnode *vn)
{
	nfs_vput(vn);
}

static void
nfs_read_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;
	u32		count;

	vn = (struct nfsnode *) req->rq_clntdata;

	if (buffer == NULL) {	/* timed out */
		nfs_vput(vn);
		return;
	}

	buffer = xdr_nfs_read_res(buffer, &vn->n_fattr, &count);
	if (nfs_check("read", vn->n_path)) {
		vn->n_flags |= N_VALIDFATTR;
	} else if (count < nfs_rsize) {
		script_error("read returned less than rsize bytes (%d)\n",
				(unsigned int) count);
	}

	nfs_notify(vn);
	nfs_vput(vn);
}

int
nfs_read(struct nfsnode *vn, u32 offset)
{
	struct nfsreq	*req;
	u32		*msg;

	if ((req = generate_request()) == NULL)
		return 0;

	/*
	printf("%08ld: requesting %08ld bytes at %08ld\n",
			offset / nfs_rsize, nfs_rsize, offset);
	 */
	req->rq_addr = &nfs_server;
	msg = xdr_nfs_read(req->rq_msg, &vn->n_fh, offset, nfs_rsize);
	req->rq_msglen = (msg - req->rq_msg) * 4;
	req->rq_callback = nfs_read_reply;
	req->rq_clntdata = vn;
	enqueue_request(req);

	return 1;
}

static void
nfs_write_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;

	if (buffer == NULL)	/* timed out */
		return;

	buffer = xdr_nfs_write_res(buffer, &vn->n_fattr);
	if (nfs_check("write", vn->n_path)) {
		vn->n_flags |= N_VALIDFATTR;
	}
	nfs_notify(vn);
	nfs_vput(vn);
}

int
nfs_write(struct nfsnode *vn, u32 offset)
{
	struct nfsreq	*req;
	u32		*msg;
	u32		count;

	if ((req = generate_request()) == NULL)
		return 0;

	count = nfs_wsize;

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_write(req->rq_msg, &vn->n_fh, offset, count);
	memset(msg, 0, count);
	req->rq_msglen = (msg - req->rq_msg) * 4 + count;
	req->rq_callback = nfs_write_reply;
	req->rq_clntdata = vn;
	enqueue_request(req);

	return 1;
}
