/*
 * inode.c	getattr/setattr for nfsbench
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <string.h>
#include "nfsbench.h"
#include "fs/nfs.h"

void
nfs_clear_sattr(struct nfs_sattr *sattr)
{
	memset(sattr, ~0, sizeof(*sattr));
}

static void
nfs_getattr_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_getattr_res(buffer, &vn->n_fattr);
	if (nfs_check("getattr", vn->n_path)) {
		vn->n_flags |= N_VALIDFATTR;
		vn->n_type = vn->n_fattr.type;
	}
	nfs_notify(vn);
	nfs_vput(vn);
}

int
nfs_getattr(struct nfsnode *vn)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request())) {
		nfs_vput(vn);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_getattr(req->rq_msg, &vn->n_fh);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_getattr_reply;
	req->rq_clntdata = vn;

	enqueue_request(req);

	return 1;
}

static void
nfs_setattr_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_setattr_res(buffer, &vn->n_fattr);
	if (nfs_check("setattr", vn->n_path)) {
		vn->n_flags |= N_VALIDFATTR;
		vn->n_type = vn->n_fattr.type;
	}
	nfs_notify(vn);
	nfs_vput(vn);
}

int
nfs_setattr(struct nfsnode *vn, struct nfs_sattr *sattr)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request())) {
		nfs_vput(vn);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_setattr(req->rq_msg, &vn->n_fh, sattr);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_setattr_reply;
	req->rq_clntdata = vn;

	enqueue_request(req);

	return 1;
}

int
nfs_truncate(struct nfsnode *vn, int len)
{
	struct nfs_sattr sattr;

	nfs_clear_sattr(&sattr);
	sattr.size = 0;

	biod_sync(nfs_setattr(vn, &sattr));

	if (!(vn->n_flags & N_VALIDFATTR) || vn->n_fattr.size) {
		script_error("couldn't truncate %s\n", vn->n_path);
		return 0;
	}
	nfs_notify(vn);
	return 1;
}

static void
nfs_statfs_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;
	nfs_fstat	stat;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_statfs_res(buffer, &stat);
	nfs_check("statfs", vn->n_path);
	nfs_notify(vn);
	nfs_vput(vn);
}

int
nfs_statfs(struct nfsnode *vn, nfs_fstat *statp)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request())) {
		nfs_vput(vn);
		return 0;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_statfs(req->rq_msg, &vn->n_fh);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_statfs_reply;
	req->rq_clntdata = vn;

	enqueue_request(req);

	return 1;
}


/* This didn't fit anywhere, so why not here */
static void
nfs_null_reply(struct nfsreq *req, u32 *buffer, int len)
{
	xdr_nfs_null(buffer);
	nfs_check("null", "");
}

int
nfs_null(u32 payload)
{
	struct nfsreq	*req;
	u32		*msg;

	if (!(req = generate_request()))
		return 0;

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_null(req->rq_msg);
	req->rq_msglen = ((msg - req->rq_msg) << 2) + payload;
	req->rq_callback = nfs_null_reply;

	enqueue_request(req);

	return 1;
}

