/*
 * namei.c	namei procedures for nfsbench
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <string.h>
#include "nfsbench.h"
#include "fs/nfs.h"
#include "fs/mount.h"

static struct nfsnode *		nfsroot;

struct nfsnode *
nfs_getroot(void)
{
	if (!nfsroot) {
		nfsroot = (struct nfsnode *) xmalloc(sizeof(*nfsroot));
		nfsroot->n_name = strdup("/");
		nfsroot->n_path = strdup("/");
		nfsroot->n_count = 1;

		if (!mount_mnt(nfs_root, nfsroot)) {
			script_error("failed to mount %s\n", nfs_root);
			return NULL;
		}

		biod_sync(nfs_getattr(nfsroot));

		if (!(nfsroot->n_flags & N_VALIDFATTR)) {
			script_error("couldn't get attributes of %s\n",
					nfs_root);
			return NULL;
		}
	}

	nfsroot->n_count++;
	return nfsroot;
}

static void
nfs_lookup_reply(struct nfsreq *req, u32 *buffer, int len)
{
	struct nfsnode	*vn;

	vn = (struct nfsnode *) req->rq_clntdata;
	xdr_nfs_lookup_res(buffer, &vn->n_fh, &vn->n_fattr);
	if (nfs_check("lookup", vn->n_path)) {
		vn->n_flags |= N_VALIDFH | N_VALIDFATTR;
		vn->n_type = vn->n_fattr.type;
	}
	nfs_notify(vn);

	/* don't vput; lookup establishes a control on the vnode */
	/* nfs_vput(vn); */
}

struct nfsnode *
nfs_lookup_cache(struct nfsnode *vn, char *name)
{
	struct nfsnode	*ch;

	if (!*name || !strcmp(name, "."))
		return vn;
	if (!strcmp(name, "..")) {
		if (!vn->n_parent)
			return vn;
		ch = vn->n_parent;
		ch->n_count++;
		nfs_vput(vn);
		return ch;
	}
	if (vn->n_flags & N_DELETED) {
		script_error("lookup on deleted directory inode %s\n",
				vn->n_path);
		nfs_vput(vn);
		return NULL;
	}
	for (ch = vn->n_children; ch; ch = ch->n_next) {
		if (!strcmp(ch->n_name, name)) {
			nfs_vput(vn);
			ch->n_count++;
			return ch;
		}
	}

	ch = (struct nfsnode *) xmalloc(sizeof(*ch));
	ch->n_path = (char *) xmalloc(strlen(vn->n_path) + strlen(name) + 2);
	strcpy(ch->n_path, vn->n_path);
	strcat(ch->n_path, "/");
	strcat(ch->n_path, name);

	ch->n_name = xstrdup(name);
	ch->n_parent = vn;
	ch->n_next = vn->n_children;
	vn->n_children = ch;
	ch->n_count++;
	nfs_vput(vn);

	return ch;
}

struct nfsnode *
nfs_lookup(struct nfsnode *vn, char *name)
{
	struct nfsnode	*ch;
	struct nfsreq	*req;
	u32		*msg;

	vn->n_count++;
	if ((ch = nfs_lookup_cache(vn, name)) == NULL) {
		nfs_vput(vn);
		return NULL;
	}

	if ((ch->n_flags & N_VALIDFH) && (ch->n_flags & N_VALIDFATTR)) {
		nfs_vput(vn);
		return ch;
	}

	if (!(req = generate_request())) {
		nfs_vput(ch);
		nfs_vput(vn);
		return NULL;
	}

	req->rq_addr = &nfs_server;
	msg = xdr_nfs_lookup(req->rq_msg, &vn->n_fh, ch->n_name);
	req->rq_msglen = (msg - req->rq_msg) << 2;
	req->rq_callback = nfs_lookup_reply;
	req->rq_clntdata = ch;

	enqueue_request(req);

	nfs_vput(vn);
	return ch;
}

struct nfsnode *
nfs_namei(struct nfsnode *vn, char *path)
{
	char	*sp, *cp, c;

	path = xstrdup(path);
	sp = path;
	while (*sp && vn) {
		while (*sp == '/')
			sp++;
		cp = sp;
		while (*sp && *sp != '/')
			sp++;
		c = *sp;
		*sp = '\0';

		biod_sync(vn = nfs_lookup(vn, cp));
		if (!(vn->n_flags & N_VALIDFH)) {
			vn = NULL;
			break;
		}
		*sp = c;
	}

	xfree(path);
	return vn;
}

struct nfsnode *
nfs_parent(struct nfsnode *vn)
{
	if (vn->n_parent)
		vn = vn->n_parent;
	else
		vn = nfsroot;
	vn->n_count++;
	return vn;
}

void
#ifdef DEBUG
_nfs_vput(struct nfsnode *vn, char *fn, int ln)
#else
nfs_vput(struct nfsnode *vn)
#endif
{
	/* printf("%s:%d: vput(%s)\n", fn, ln, vn->n_path); */
	/*
	if (!(vn->n_flags & N_VALIDFH))
		script_error("invalid fh: %s\n", vn->n_path);
	 */
	if (vn->n_count <= 0) {
#ifdef DEBUG
		script_error("nfs_vput: (%s: %d) n_count 0 in %s\n",
					fn, ln, vn->n_path);
#else
		script_error("nfs_vput: n_count 0 in %s\n", vn->n_path);
#endif
		vn->n_count = 1;
	}
	if (--(vn->n_count) > 0 || !(vn->n_flags & N_DELETED))
		return;
	nfs_vremove(vn);
}

void
nfs_vremove(struct nfsnode *vn)
{
	struct nfsnode	*pt, **vp;

	if ((pt = vn->n_parent) == NULL)
		return;

	if (vn->n_children) {
		script_error("vremove: %s still has children\n", vn->n_path);
		return;
	}

	for (vp = &pt->n_children; *vp; vp = &((*vp)->n_next)) {
		if (*vp == vn) {
			*vp = vn->n_next;
			break;
		}
	}
	vn->n_flags |= N_DELETED;

	if (vn->n_count > 1)
		return;

	/*
	if ((pt->n_flags & N_DELETED) && vn->n_parent)
		nfs_vremove(vn->n_parent);
	 */

	xfree(vn->n_name);
	xfree(vn->n_path);
	xfree(vn->n_symlink);
	xfree(vn);
}
