/*
 * nfs.h	NFS file system declarations.
 *
 *		The naming conventions aren't what I'd call unambiguous.
 *		There are high-level routines similar to the VFS which
 *		work synchronously (i.e. the result is there when they
 *		return), and low-level rpc functions that do nothing
 *		but queue the call.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#ifndef FS_NFS_H
#define FS_NFS_H

#include "xdr/nfs.h"

#define N_VALIDFH	0x0001
#define N_VALIDFATTR	0x0002
#define N_DELETED	0x0004
#define N_CHAINED	0x0008
#define N_LOCKED	0x0010

struct nfsnode {
	struct nfsnode *	n_next;
	struct nfsnode *	n_children;
	struct nfsnode *	n_parent;
	struct nfsnode *	n_chain;
	char *			n_path;
	char *			n_name;
	char *			n_symlink;
	unsigned short		n_flags;
	unsigned short		n_type;
	int			n_count;
	struct nfs_fh		n_fh;
	struct nfs_fattr	n_fattr;

	/* Notifier function called after NFS return */
	void			(*n_notify)(struct nfsnode *);
	void *			n_clntdata;
};

struct nfsdir {
	struct nfsnode *	d_vnode;
	u32			d_cookie;
	struct nfsdir *		d_chain;
	unsigned short		d_eof,
				d_pend;

	/* Notifier function called after readdir return */
	void			(*d_notify)(struct nfsdir *);
	void *			d_clntdata;
};

struct nfsnode *	nfs_getroot(void);
struct nfsnode *	nfs_lookup_cache(struct nfsnode *, char *);
struct nfsnode *	nfs_namei(struct nfsnode *, char *path);
struct nfsnode *	nfs_parent(struct nfsnode *);
#ifdef DEBUG
void			_nfs_vput(struct nfsnode *, char *fn, int ln);
#define nfs_vput(vn)	_nfs_vput(vn, __FILE__, __LINE__);
#else
void			nfs_vput(struct nfsnode *);
#endif
void			nfs_vremove(struct nfsnode *);
struct nfsnode *	nfs_open(char *name, unsigned int mode);
struct nfsnode *	nfs_creat(char *name, unsigned int mode);
void			nfs_close(struct nfsnode *file);
struct nfsdir *		nfs_opendir(struct nfsnode *);
void			nfs_closedir(struct nfsdir *);
int			nfs_readdir(struct nfsdir *);
void			nfs_clear_sattr(struct nfs_sattr *);

int			nfs_null(u32 payload);
int			nfs_getattr(struct nfsnode *);
int			nfs_setattr(struct nfsnode *, struct nfs_sattr *);
struct nfsnode *	nfs_lookup(struct nfsnode *, char *);
int			nfs_readlink(struct nfsnode *);
struct nfsnode *	nfs_create(struct nfsnode *, char *, struct nfs_sattr *);
struct nfsnode *	nfs_mkdir(struct nfsnode *, char *, struct nfs_sattr *);
int			nfs_remove(struct nfsnode *, char *);
int			nfs_rmdir(struct nfsnode *, char *);
int			nfs_read(struct nfsnode *, u32 offset);
int			nfs_write(struct nfsnode *, u32 offset);
struct nfsnode *	nfs_symlink(struct nfsnode *, char *, char *, 
						struct nfs_sattr *);
struct nfsnode *	nfs_rename(struct nfsnode *, char *,
						struct nfsnode *, char *);
struct nfsnode *	nfs_link(struct nfsnode *, struct nfsnode *, char *);
int			nfs_truncate(struct nfsnode *vn, int len);
int			nfs_statfs(struct nfsnode *vn, nfs_fstat *);

int			nfs_check(char *op, char *name);

#define dir_notify(dir)				\
	do {					\
		if ((dir)->d_notify)		\
			(dir)->d_notify(dir);	\
	} while (0)

#define nfs_notify(vn)				\
	do {					\
		if ((vn)->n_notify)		\
			(vn)->n_notify(vn);	\
	} while (0)

#endif /* FS_NFS_H */
