/*
 * find_script.c	Recursive find of a dir hierarchy
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */
#include <sys/types.h>
#include <sys/param.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "nfsbench.h"
#include "xgetopt.h"
#include "xdr/nfs.h"
#include "fs/nfs.h"

struct context;

static void *	find_init(int argc, char **argv);
static int	find(struct context *);
static void	find_finish(struct context *);
static void	find_children(struct context *, struct nfsnode *);
static void	sched_action(struct context *, struct nfsnode *);
static int	do_readdir(struct context *ctxt);
static int	do_lookup(struct context *ctxt);
static int	do_remove(struct context *ctxt);

static action_fn_t	actions[] = {
	(action_fn_t) find,
};

struct script	find_script = {
	"find",
	(scinit_fn_t) find_init,
	(scdone_fn_t) find_finish,
	nrof(actions),
	actions
};

struct context {
	struct nfsdir *		readdir;
	struct nfsnode *	lookup;
	struct nfsnode *	remove;
	struct nfsnode *	pending;
	int			count;
	int			silent;
	int			op;
};

enum { FOP_NOP, FOP_RM };

#define DCHAIN(ctxt, chain, dp)		\
	dp->d_chain = chain;		\
	chain = dp;			\
	ctxt->count++;
#define DUNCHAIN(ctxt, chp, dp)		\
	*chp = dp->d_chain;		\
	ctxt->count--;
#define FCHAIN(ctxt, chain, vn)		\
	vn->n_chain = chain;		\
	chain = vn;			\
	vn->n_flags |= N_CHAINED;	\
	ctxt->count++;
#define FUNCHAIN(ctxt, chp, vn)		\
	*chp = vn->n_chain;		\
	vn->n_flags &= ~N_CHAINED;	\
	ctxt->count--;

/* #define TRACKNAME			"peb21" */
#ifdef TRACKNAME
#define TRACK(vn)			\
	if (!strcmp(vn->n_name, TRACKNAME)) \
		printf("XXX[%d]: %s\n", vn->n_count, vn->n_path);
#else
#define TRACK(vn)			do { } while (0)
#endif

/*
 * READDIR completed successfully
 * One control on input, one on output.
 */
static void
readdir_notify(struct nfsdir *dir)
{
	struct context	*ctxt = (struct context *) dir->d_clntdata;
	struct nfsnode	*vn = dir->d_vnode;

	if (dir->d_eof) {
		/* keep counter on directory */
		vn->n_count++;
		nfs_closedir(dir);
		TRACK(vn);
		find_children(ctxt, vn);
		if (!vn->n_children)
			sched_action(ctxt, vn);
		return;
	}

	/* Read more entries */
	DCHAIN(ctxt, ctxt->readdir, dir);
	TRACK(vn);
}

/*
 * Schedule readdir.
 * vn	One control on input.
 *	opendir doesn't esatblish additional control.
 */
static void
sched_readdir(struct context *ctxt, struct nfsnode *vn)
{
	struct nfsdir	*dir;

	dir = nfs_opendir(vn);
	dir->d_clntdata = ctxt;
	dir->d_notify   = readdir_notify;
	DCHAIN(ctxt, ctxt->readdir, dir);
	TRACK(vn);
	do_readdir(ctxt);
}

/*
 * Perform readdir call.
 * vn	One control on input.
 *	On EOF, release one control, otherwise no change.
 */
static int
do_readdir(struct context *ctxt)
{
	struct nfsdir	*dir, **chp;
	struct nfsnode	*vn;
	int		dirop = 0;

	chp = &ctxt->readdir;
	while ((dir = *chp) != NULL) {
		vn = dir->d_vnode;

		if (dir->d_eof) {
			script_error("EOF dir in readdir chain?\n");
			DUNCHAIN(ctxt, chp, dir);
			nfs_closedir(dir);
			continue;
		}

		if (!nfs_readdir(dir))
			break;
		DUNCHAIN(ctxt, chp, dir);
		TRACK(vn);
		/* find_children(ctxt, vn); */
		dirop++;
	}
	return dirop;
}

/*
 * Find children.
 * Establish one additonal control on each child that hasn't been scheduled
 * yet.
 */
static void
find_children(struct context *ctxt, struct nfsnode *vn)
{
	struct nfsnode	*ch;

	for (ch = vn->n_children; ch; ch = ch->n_next) {
		if (ch->n_flags & (N_LOCKED|N_CHAINED))
			continue;

		/* Add one control */
		ch->n_count++;

		TRACK(ch);
		if (ch->n_flags & N_VALIDFATTR) {
			if (ch->n_type == NFDIR) {
				sched_readdir(ctxt, ch);
			} else {
				sched_action(ctxt, ch);
			}
		} else {
			FCHAIN(ctxt, ctxt->lookup, ch);
		}
	}
	if (ctxt->lookup)
		do_lookup(ctxt);
}

/*
 * LOOKUP call completed.
 * vn has one control on input. If the vnode should be further acted upon,
 * it is retained, otherwise it is released.
 */
static void
lookup_notify(struct nfsnode *vn)
{
	struct context	*ctxt = (struct context *) vn->n_clntdata;

	TRACK(vn);
	vn->n_notify = NULL;
	if (!(vn->n_flags & N_VALIDFATTR)) {
		script_error("lookup for %s failed\n", vn->n_path);
		return;
	}
	vn->n_flags &= ~N_LOCKED;

	if (vn->n_type == NFDIR) {
		sched_readdir(ctxt, vn);
	} else {
		sched_action(ctxt, vn);
	}
}

/*
 * Perform lookup call.
 * vn has one control on input, which is retained
 */
static int
do_lookup(struct context *ctxt)
{
	struct nfsnode	*vn;
	int		dirop = 0;

	while ((vn = ctxt->lookup) != NULL) {
		if (!nfs_lookup(nfs_parent(vn), vn->n_name))
			break;
		FUNCHAIN(ctxt, &ctxt->lookup, vn);
		TRACK(vn);
		vn->n_flags   |= N_LOCKED;
		vn->n_notify   = lookup_notify;
		vn->n_clntdata = ctxt;
		dirop++;
	}
	return dirop;
}

/*
 * Schedule the vnode for removal etc.
 * One control on input, released when no action, otherwise retained
 */
static void
sched_action(struct context *ctxt, struct nfsnode *vn)
{
	TRACK(vn);

	switch (ctxt->op) {
	case FOP_RM:
		if (vn->n_parent != NULL) {
			FCHAIN(ctxt, ctxt->remove, vn);
			do_remove(ctxt);
			return;
		}
		break;
	default:
		nfs_vput(vn);
	}
}

/*
 * File removed. The notifier is called before deleting the vnode.
 *
 * If this was the last file in the given directory, schedule
 * the parent directory for removal. Note that nfs_parent establishes
 * one control on the parent dir.
 */
static void
remove_notify(struct nfsnode *vn)
{
	struct context	*ctxt = (struct context *) vn->n_clntdata;

	TRACK(vn);
	vn->n_notify = NULL;

	if (vn->n_parent->n_children == vn && !vn->n_next)
		sched_action(ctxt, vn->n_parent);

	nfs_vput(vn);
	ctxt->count--;
}

/*
 * Schedule remove call.
 * Each node has one control on input, one on output
 * The parent's refcount remains unchanged, too.
 */
static int
do_remove(struct context *ctxt)
{
	struct nfsnode	*vn, **chp;
	int		dirop = 0;

	chp = &ctxt->remove;
	while ((vn = *chp) != NULL) {
		vn->n_notify   = remove_notify;
		vn->n_clntdata = ctxt;
		if (vn->n_type != NFDIR) {
			if (!nfs_remove(nfs_parent(vn), vn->n_name))
				break;
		} else {
			if (!nfs_rmdir(nfs_parent(vn), vn->n_name))
				break;
		}
		FUNCHAIN(ctxt, chp, vn);
		nfs_vput(vn);
		TRACK(vn);
		ctxt->count++;
		dirop++;
	}

	return dirop;
}

static int
find(struct context *ctxt)
{
	int	ops, total = 0;

	do {
		ops = do_readdir(ctxt) +
		      do_lookup(ctxt) +
		      do_remove(ctxt);
		total += ops;
	} while (ops);

	if (!ctxt->count)
		return AR_OK;

	if (!total)
		biod_flush();

	return AR_AGAIN;
}

/*
 * Options understood by find script
 */
static char *	path;
static int	silent;
static int	delete;
static struct xoption	longopts[] = {
 { "path",	'p',	OPT_STRING,	NULL,	&path,
					"Directory",		},
 { "help",	'h',	OPT_HELP,	NULL,	NULL,
					"Print this message"	},
 { "silent",	's',	OPT_TRUE,	NULL,	&silent,
					"Don't print stats"	},
 { "rm",	'r',	OPT_TRUE,	NULL,	&delete,
					"Remove files"		},

 { NULL,	0,	OPT_END,	NULL,	NULL,	NULL	}
};

/*
 * Initialize find script
 */
static void *
find_init(int argc, char **argv)
{
	struct context	*ctxt;
	struct nfsnode	*vn;
	int		idx;

	path = "/";
	silent = 0;

	if (!xgetopt(argc, argv, longopts, &idx))
		return 0;

	if (!(vn = nfs_getroot()) || !(vn = nfs_namei(vn, path))
	 || !(vn->n_flags & N_VALIDFH))
		return NULL;

	ctxt = (struct context *) xmalloc(sizeof(*ctxt));
	ctxt->silent = silent;
	ctxt->count = 0;

	if (delete) {
		ctxt->op = FOP_RM;
	} else {
		ctxt->op = FOP_NOP;
	}

	setbuf(stdout, NULL);
	sched_readdir(ctxt, vn);
	clear_stats();

	return ctxt;
}

static void
find_finish(struct context *ctxt)
{
	if (ctxt) {
		if (!ctxt->silent)
			print_stats();
		xfree(ctxt);
	}
}
