/*
 * nfspebbles.c
 *		Determine avg. number of NFS operations per second.
 *
 * This was inspired by a description of nhfsstone by Legato, Inc.,
 * and Sun's nfsstone. The difference is that those programs go through
 * the VFS and merely estimate the number and mix of NFS calls actually
 * placed to the server.
 * 
 * Oh, and in case you didn't notice. This code is a complete mess.
 *
 * FIXME: make sure vnode->n_count is always maintained properly.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <sys/param.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "nfsbench.h"
#include "xgetopt.h"
#include "fs/nfs.h"

#define dprintf	if (0) printf
#ifndef MAX
# define MAX(a, b)	((a) > (b)? (a) : (b))
#endif

struct context;

static int	gendirs(struct context *ctxt);
static int	genfiles(struct context *ctxt);
static int	gensymlks(struct context *ctxt);
static int	writefiles(struct context *ctxt);
static int	ignition(struct context *ctxt);
static int	pebbles(struct context *);
static int	pebbles_do(struct nfsreq *, struct context *, int);
static void *	pebbles_init(int argc, char **argv);
static void	pebbles_finish(struct context *);

static action_fn_t	actions[] = {
	(action_fn_t) gendirs,
	(action_fn_t) genfiles,
	(action_fn_t) gendirs,
	(action_fn_t) writefiles,
	(action_fn_t) gensymlks,
	(action_fn_t) ignition,
	(action_fn_t) pebbles,
};

struct script	pebbles_script = {
	"nfspebbles",
	(scinit_fn_t) pebbles_init,
	(scdone_fn_t) pebbles_finish,
	nrof(actions),
	actions
};

/*
 * Different NFS call mixtures (percent)
 *
 *	null	getattr	setattr	root	lookup	readlink
 *	read	wcache	write	create	remove	rename
 *	link	symlink	mkdir	rmdir	readdir	fsstat
 *
 * It's interesting to see that both mixes create more files than
 * they remove :-)
 */
/* sundev:	Taken from paper [1] */
static int	sundev[18] = {
	0,	5,	0,	0,	50,	7,
	30,	0,	3,	1,	0,	0,
	0,	0,	0,	0,	0,	0
};
/* legato:	the `legato' mix from nhfsstone, according to [4] */
static int	legato[18] = {
	0,	13,	1,	0,	34,	8,
	22,	0,	15,	2,	1,	0,
	0,	0,	0,	0,	3,	1
};
/* debug:	a test mix for debugging (all operations tested) */
static int	debug[18] = {
	0,	1,	1,	0,	1,	1,
	1,	0,	1,	1,	1,	1,
	1,	1,	1,	1,	1,	1
};

struct context {
	int		maxcalls;
	int *		mix;

	int		nrfiles;
	int		nrdirs;
	int		nrsyms;
	int		fsize;
	int		numobjs;
	struct fsobj *	curobj;
	u32		cursize;

	int		nrcalls;
	int		calls[18];
	int		nrops;
	int		idxop;
	int		ops[100];

	struct fsobj *	dirs;
	struct fsobj *	syms;
	struct fsobj *	files;
};

struct fsobj {
	struct fsobj *	next;
	struct fsobj *	prev;
	struct nfsnode *vnode;
	struct nfsdir *	dir;
};

static int	peb_op_null(struct context *, struct fsobj **, char *);
static int	peb_op_getattr(struct context *, struct fsobj **, char *);
static int	peb_op_setattr(struct context *, struct fsobj **, char *);
static int	peb_op_root(struct context *, struct fsobj **, char *);
static int	peb_op_lookup(struct context *, struct fsobj **, char *);
static int	peb_op_readlink(struct context *, struct fsobj **, char *);
static int	peb_op_read(struct context *, struct fsobj **, char *);
static int	peb_op_wcache(struct context *, struct fsobj **, char *);
static int	peb_op_write(struct context *, struct fsobj **, char *);
static int	peb_op_create(struct context *, struct fsobj **, char *);
static int	peb_op_remove(struct context *, struct fsobj **, char *);
static int	peb_op_rename(struct context *, struct fsobj **, char *);
static int	peb_op_link(struct context *, struct fsobj **, char *);
static int	peb_op_symlink(struct context *, struct fsobj **, char *);
static int	peb_op_mkdir(struct context *, struct fsobj **, char *);
static int	peb_op_rmdir(struct context *, struct fsobj **, char *);
static int	peb_op_readdir(struct context *, struct fsobj **, char *);
static int	peb_op_fsstat(struct context *, struct fsobj **, char *);

typedef int	(*bebop_fn_t)(struct context *, struct fsobj **, char *);
static bebop_fn_t	bebops[18] = {
	peb_op_null,	peb_op_getattr,	peb_op_setattr,	peb_op_root,
	peb_op_lookup,	peb_op_readlink,peb_op_read,	peb_op_wcache,
	peb_op_write,	peb_op_create,	peb_op_remove,	peb_op_rename,
	peb_op_link,	peb_op_symlink,	peb_op_mkdir,	peb_op_rmdir,
	peb_op_readdir,	peb_op_fsstat
};
/*
 * Different types of args: none, any (file or dir), file, dir,
 * dir+name, file+dir+name (for link/symlink/rename), symlink (for readlink)
 */
enum {	NARG, AARG, FARG, DARG, DNARG, LARG, SARG };
static int	opargs[18] = {
	NARG,		AARG,		AARG,		NARG,
	AARG,		SARG,		FARG,		NARG,
	FARG,		DNARG,		FARG,		LARG,
	LARG,		DNARG,		DNARG,		DARG,
	DARG,		AARG
};


static char *
getname(void)
{
	static char	buffer[12];
	static int	fcnt = 1;

	sprintf(buffer, "peb%d", fcnt++);
	return buffer;
}

static void
finsert(struct fsobj **head, struct fsobj *obj)
{
	struct fsobj	*ins;
	int		i;

	if ((ins = *head) != NULL) {
		obj->next = ins;
		obj->prev = ins->prev;
		ins->prev = obj;
		obj->prev->next = obj;
	} else {
		obj->next = obj->prev = obj;
	}

	/* Jumble file list a bit */
	for (i = xtime % 3; i; i--)
		obj = obj->next;

	*head = obj;
}

static void
fremove(struct fsobj **head, struct fsobj *obj)
{
	struct fsobj	*next = obj->next, *prev = obj->prev;

	if (next == obj) {
		*head = NULL;
	} else {
		prev->next = next;
		next->prev = prev;
		*head = next;
	}
}

static struct fsobj *
fcreate(struct fsobj **head, struct nfsnode *vn)
{
	struct fsobj	*obj;

	obj = (struct fsobj *) xmalloc(sizeof(*obj));
	obj->vnode = vn;
	finsert(head, obj);
	return obj;
}

static void
fdelete(struct fsobj **head, struct fsobj *obj)
{
	fremove(head, obj);
	nfs_vput(obj->vnode);
	xfree(obj);
}

#define OP_DO(stmt)		\
    do {			\
	if (!(stmt))		\
		return 0;	\
	op_lock(vn);		\
	return 1;		\
    } while (0)

static void
op_notify(struct nfsnode *vn)
{
	dprintf("X %3d %s\n", vn->n_count, vn->n_path);
	vn->n_notify = NULL;
	vn->n_flags &= ~N_LOCKED;
}

static void
op_lock(struct nfsnode *vn)
{
	vn->n_notify = op_notify;
	vn->n_flags |= N_LOCKED;
}

static struct fsobj *
getobj(struct fsobj **head, char *what, int op, int delete)
{
	struct fsobj	*obj;
	struct nfsnode	*vn;
	int		i;

	if (!(obj = *head)) {
		script_error("ran out of %s!\n", what);
		exit (2);
	}

	/* Jumble file list a bit */
	for (i = xtime % 3; i; i--)
		obj = obj->next;

	while (1) {
		vn = obj->vnode;
		if (!(vn->n_flags & N_LOCKED)) {
			if ((vn->n_flags & N_VALIDFH) || op == NFSPROC_LOOKUP)
				break;
		}
		if ((obj = obj->next) == *head)
			return NULL;
	}

	while (delete) {
		if (!vn->n_children) {
			if (obj->dir && vn->n_count == 1) {
				nfs_closedir(obj->dir);
			}
			obj->dir = NULL;
			if (!vn->n_count)
				break;
		}
		obj = obj->next;
		vn = obj->vnode;
		if (obj == *head) {
			/* script_error("no deletable %s found\n", what); */
			return NULL;
		}
	}

	vn->n_count++;
#if 0
	printf("%c %3d %s\n",
		(vn->n_type == NFDIR)? 'D' : 
		((vn->n_type == NFREG)? 'F' : 'S'),
		vn->n_count, vn->n_path);
#endif
	return obj;
}

static struct fsobj *
getfile(struct context *ctxt, int op)
{
	return getobj(&ctxt->files, "files", op, op == NFSPROC_REMOVE);
}

static struct fsobj *
getdir(struct context *ctxt, int op)
{
	return getobj(&ctxt->dirs, "directories", op, op == NFSPROC_RMDIR);
}

static struct fsobj *
getsymlk(struct context *ctxt, int op)
{
	return getobj(&ctxt->syms, "symlinks", op, op == NFSPROC_REMOVE);
}

static int
gendirs(struct context *ctxt)
{
	struct nfsnode	*vn;
	char		*name;
	struct nfs_sattr sattr;

	nfs_clear_sattr(&sattr);
	sattr.mode = 0777;

	while (1) {
		if (!ctxt->dirs) {
			if (!(vn = nfs_getroot())) {
				script_error("couldn't get fs root\n");
				return AR_ABORT;
			}
		} else {
			vn = getdir(ctxt, NFSPROC_MKDIR)->vnode;
		}
		if (!(vn->n_flags & N_VALIDFH))
			biod_flush();
		name = getname();
		biod_sync(vn = nfs_mkdir(vn, name, &sattr));
		if (!vn || !(vn->n_flags & N_VALIDFH))
			return AR_AGAIN;
		fcreate(&ctxt->dirs, vn);
		if (++(ctxt->numobjs) > ctxt->nrdirs)
			break;
	}

	ctxt->numobjs = 0;
	return AR_OK;
}

static int
genfiles(struct context *ctxt)
{
	struct nfsnode	*vn;
	char		*name;
	struct nfs_sattr sattr;

	nfs_clear_sattr(&sattr);
	sattr.mode = 0666;

	while (1) {
		vn = getdir(ctxt, NFSPROC_CREATE)->vnode;
		name = getname();
#if 0
		biod_sync(vn = nfs_create(vn, name, &sattr));
		if (!vn || !(vn->n_flags & N_VALIDFH))
			return AR_AGAIN;
#else
		if (!(vn = nfs_create(vn, name, &sattr)))
			return AR_AGAIN;
#endif
		fcreate(&ctxt->files, vn);
		if (++(ctxt->numobjs) > ctxt->nrfiles)
			break;
	}

	ctxt->curobj = ctxt->files;
	ctxt->cursize = 0;
	ctxt->numobjs = 0;
	return AR_OK;
}

static int
gensymlks(struct context *ctxt)
{
	struct nfsnode	*vn;
	char		*name;
	struct nfs_sattr sattr;
	struct fsobj	*obj;
	int		drain = 0;

	nfs_clear_sattr(&sattr);
	sattr.mode = 0666;

	while (ctxt->numobjs <= ctxt->nrsyms) {
		vn = getdir(ctxt, NFSPROC_SYMLINK)->vnode;
		name = getname();
		if (!(vn = nfs_symlink(vn, name, "../yabba/whap", &sattr)))
			return AR_AGAIN;
		fcreate(&ctxt->syms, vn);
		ctxt->numobjs++;
		drain++;
	}

	if (drain)
		biod_drain();

	obj = ctxt->syms;
	while (!((vn = obj->vnode)->n_flags & (N_VALIDFH|N_LOCKED))) {
		if (!nfs_lookup(nfs_parent(vn), vn->n_name))
			return AR_AGAIN;
		vn->n_notify = op_notify;
		vn->n_flags |= N_LOCKED;
		obj = obj->next;
	}

	ctxt->curobj = ctxt->syms;
	ctxt->cursize = 0;
	ctxt->numobjs = 0;
	return AR_OK;
}

static int
writefiles(struct context *ctxt)
{
	struct fsobj	*obj;
	struct nfsnode	*vn;

	while (1) {
		obj = ctxt->curobj;
		vn = obj->vnode;
		if (!(vn->n_flags & N_VALIDFATTR)) {
			script_error("create of %s failed\n", vn->n_path);
			fdelete(&ctxt->files, obj);
			return AR_ERROR;
		}
		while (ctxt->cursize < ctxt->fsize) {
			vn->n_count++;
			if (!nfs_write(vn, ctxt->cursize))
				return AR_AGAIN;
			ctxt->cursize += nfs_wsize;
		}
		ctxt->cursize = 0;
		if ((ctxt->curobj = obj->next) == ctxt->files)
			break;
	}

	return AR_OK;
}

static int
ignition(struct context *ctxt)
{
	biod_flush();
	printf("Generating random requests...\n");
	clear_stats();
	return AR_OK;
}

static int
pebbles(struct context *ctxt)
{
	int	op, i;

	while(1) {
		do {
			i = rand() % ctxt->nrops;
		} while (!ctxt->mix[(op = ctxt->ops[i])]);

		if (!pebbles_do(NULL, ctxt, op))
			break;

		ctxt->calls[op]++;
		if (++(ctxt->nrcalls) >= ctxt->maxcalls)
			return AR_OK;
	}
	return AR_AGAIN;
}

static int
pebbles_do(struct nfsreq *req, struct context *ctxt, int op)
{
	struct fsobj	*argv[2];
	int		atype = opargs[op], argc = 0, i;
	char		*name = NULL;

	/* printf("%s:\n", xdr_nfs_name(op)); */
	if (atype == AARG) 
		atype = (xtime & 1)? FARG : DARG;
	if (atype == FARG || atype == LARG)
		argv[argc++] = getfile(ctxt, op);
	if (atype == DARG || atype == LARG || atype == DNARG)
		argv[argc++] = getdir(ctxt, op);
	if (atype == SARG)
		argv[argc++] = getsymlk(ctxt, op);
	if (atype == DNARG || atype == LARG)
		name = getname();
	for (i = 0; i < argc; i++)
		if (!argv[i])
			goto fail;

	return (bebops[op])(ctxt, argv, name);

fail:
	dprintf("releasing args for op %s\n", xdr_nfs_name(op));
	while (argc--) {
		if (argv[argc])
			nfs_vput(argv[argc]->vnode);
	}
	return 0;
}

static int
peb_op_null(struct context *ctxt, struct fsobj **argv, char *name)
{
	dprintf("null()\n");
	return nfs_null(0);
}

static int
peb_op_getattr(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("getattr(%s)\n", argv[0]->vnode->n_path);
	OP_DO(nfs_getattr(vn));
}

static int
peb_op_setattr(struct context *ctxt, struct fsobj **argv, char *name)
{
	static struct nfs_sattr	sattr;
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("setattr(%s)\n", vn->n_path);
	nfs_clear_sattr(&sattr);
	sattr.mtime.tv_sec = xtime / HZ;
	sattr.mtime.tv_usec = 0;

	OP_DO(nfs_setattr(vn, &sattr));
}

static int
peb_op_root(struct context *ctxt, struct fsobj **argv, char *name)
{
	return 0;
}

static void
lookup_notify(struct nfsnode *vn)
{
	op_notify(vn);
	nfs_vput(vn);
}

static int
peb_op_lookup(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;
	struct nfsnode	*ch;

	dprintf("lookup(%s)\n", vn->n_path);
	vn->n_flags &= ~N_VALIDFATTR;	/* force NFS call */

	if (!(ch = nfs_lookup(nfs_parent(vn), vn->n_name)))
		return 0;
	nfs_vput(vn);
	ch->n_notify = lookup_notify;
	ch->n_flags |= N_LOCKED;
	return 1;
}

static int
peb_op_readlink(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("readlink(%s)\n", vn->n_path);
	OP_DO(nfs_readlink(vn));
}

/* FIXME: vnode refcount seems to be be off by one after READ */
static int
peb_op_read(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;
	u32		size = vn->n_fattr.size, offset;

	dprintf("read(%s)\n", vn->n_path);
	if (size < nfs_rsize)
		return 0;
	offset = xtime % (size / nfs_rsize);
	OP_DO(nfs_read(vn, offset));
}

static int
peb_op_wcache(struct context *ctxt, struct fsobj **argv, char *name)
{
	return 0;
}

static int
peb_op_write(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;
	u32		size = vn->n_fattr.size, offset;

	dprintf("write(%s)\n", vn->n_path);
	if (size < nfs_wsize)
		offset = 0;
	else
		offset = xtime % (size / nfs_wsize);
	OP_DO(nfs_write(vn, offset));
}

static int
peb_op_create(struct context *ctxt, struct fsobj **argv, char *name)
{
	static struct nfs_sattr	sattr;
	struct nfsnode	*vn = argv[0]->vnode;
	struct nfsnode	*ch;

	dprintf("create(%s, %s)\n", vn->n_path, name);
	nfs_clear_sattr(&sattr);
	sattr.mode = 0666;
	if (!(ch = nfs_create(vn, name, &sattr)))
		return 0;
	fcreate(&ctxt->files, ch);
	op_lock(ch);
	return 1;
}

static int
peb_op_remove(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("remove(%s)\n", vn->n_path);
	if (!nfs_remove(nfs_parent(vn), vn->n_name))
		return 0;
	fdelete(&ctxt->files, argv[0]);
	/* nfs_vput(vn); */
	/* op_lock(vn); */

	return 1;
}

static int
peb_op_rename(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*fvn = argv[0]->vnode;
	struct nfsnode	*tvn = argv[1]->vnode;
	struct nfsnode	*ch;

	dprintf("rename(%s, %s/%s)\n", fvn->n_path, tvn->n_path, name);
	if (!(ch = nfs_rename(nfs_parent(fvn), fvn->n_name, tvn, name)))
		return 0;

	fdelete(&ctxt->files, argv[0]);
	fcreate(&ctxt->files, ch);
	ch->n_count++;
	op_lock(ch);

	return 1;
}

static int
peb_op_link(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*fvn = argv[0]->vnode;
	struct nfsnode	*tvn = argv[1]->vnode;
	struct nfsnode	*ch;

	dprintf("link(%s, %s/%s)\n", fvn->n_path, tvn->n_path, name);
	if (!(ch = nfs_link(fvn, tvn, name)))
		return 0;

	fcreate(&ctxt->files, ch);
	ch->n_count++;
	op_lock(ch);

	return 1;
}

static int
peb_op_symlink(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode		*vn = argv[0]->vnode;
	struct nfs_sattr	sattr;

	dprintf("symlink(%s/%s, yabba/whap)\n", vn->n_path, name);

	nfs_clear_sattr(&sattr);
	sattr.mode = 0666;

	if (!(vn = nfs_symlink(vn, name, "../yabba/whap", &sattr)))
		return 0;
	fcreate(&ctxt->syms, vn);
	op_lock(vn);
	return 1;
}

static int
peb_op_mkdir(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;
	static struct nfs_sattr	sattr;

	dprintf("mkdir(%s, %s)\n", vn->n_path, name);
	nfs_clear_sattr(&sattr);
	sattr.mode = 0777;
	if (!(vn = nfs_mkdir(vn, name, &sattr)))
		return 0;
	fcreate(&ctxt->dirs, vn);
	op_lock(vn);
	return 1;
}

static int
peb_op_rmdir(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("rmdir(%s)\n", vn->n_path);
	if (!nfs_rmdir(nfs_parent(vn), vn->n_name))
		return 0;

	fdelete(&ctxt->dirs, argv[0]);
	/* nfs_vput(vn); */
	return 1;
}

static int
peb_op_readdir(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;
	struct nfsdir	*dir;

	dprintf("readdir(%s)\n", vn->n_path);
	if (!(dir = argv[0]->dir))
		dir = nfs_opendir(vn);
	OP_DO(nfs_readdir(dir));
}

static int
peb_op_fsstat(struct context *ctxt, struct fsobj **argv, char *name)
{
	struct nfsnode	*vn = argv[0]->vnode;

	dprintf("statfs(%s)\n", vn->n_path);
	OP_DO(nfs_statfs(vn, NULL));
}

/*
 * Options understood by pebbles
 */
static char *	mixname;
static int	nrcalls;
static int	fsize;

static struct xoption	longopts[] = {
 { "help",	'h',	OPT_HELP,	NULL,	NULL,
					"Print this message"	},
 { "mix",	'm',	OPT_STRING,	NULL,	&mixname,
					"Predefined mix (legato)"},
 { "calls",	'n',	OPT_INT,	NULL,	&nrcalls,
					"Number of calls"	},
 { "size",	'f',	OPT_INT,	NULL,	&fsize,
					"File size"		},

 { NULL,	0,	OPT_END,	NULL,	NULL,	NULL	}
};

static void *
pebbles_init(int argc, char **argv)
{
	struct context	*ctxt;
	int		idx, *mix, i, j;
	int		nrfiles, nrdirs, nrsyms, delta;

	nrcalls = 0;
	mixname = NULL;

	if (!xgetopt(argc, argv, longopts, &idx))
		return NULL;

	if (mixname) {
		if (!strcmp(mixname, "sundev")) {
			mix = sundev;
		} else if (!strcmp(mixname, "legato")) {
			mix = legato;
		} else if (!strcmp(mixname, "debug")) {
			mix = debug;
		} else {
			script_error("unknown operations mix %s\n", mixname);
			return NULL;
		}
	} else {
		printf("Using legato mix\n");
		mix = legato;
	}

#if 0
	printf("Debugging: using test mix\n");
	mix = debug;
#endif

	ctxt = (struct context *) xmalloc(sizeof(*ctxt));
	if (nrcalls == 0)
		ctxt->maxcalls = 2048;
	else
		ctxt->maxcalls = nrcalls;
	if (fsize == 0)
		ctxt->fsize = 4 * MAX(nfs_rsize, nfs_wsize);
	else
		ctxt->fsize = fsize;

	ctxt->mix = mix;
	ctxt->nrops = 0;
	for (i = 0; i < 18; i++) {
		if (ctxt->nrops >= 100) {
			script_error("sum of operations > 100!\n");
			xfree(ctxt);
			return NULL;
		}
		for (j = 0; j < ctxt->mix[i]; j++)
			ctxt->ops[ctxt->nrops++] = i;
	}

	nrdirs  = 256;
	nrfiles = 256;
	if ((delta = mix[NFSPROC_RMDIR] - mix[NFSPROC_MKDIR]) >  0)
		nrdirs = (delta * ctxt->maxcalls / ctxt->nrops);
	if ((delta = mix[NFSPROC_REMOVE] - mix[NFSPROC_CREATE]) >  0)
		nrfiles = (delta * ctxt->maxcalls / ctxt->nrops);
	nrsyms = 10;

	printf("Generating %d directories, %d symlks and %d files (%d bytes)\n",
				nrdirs, nrsyms, nrfiles, ctxt->fsize);
	ctxt->nrfiles = nrfiles;
	ctxt->nrdirs = nrdirs >> 1;
	ctxt->nrsyms = nrsyms;

	clear_stats();

	return ctxt;
}

static void
pebbles_finish(struct context *ctxt)
{
	unsigned long	total = 0;
	int		i;

	if (!ctxt)
		return;

	print_stats();
	for (i = 0; i < 18; i++)
		total += ctxt->calls[i];

	printf("    call statistics:\n");
	for (i = 0; i < 18; i++) {
		if (!ctxt->mix[i])
			continue;
		printf("\t%-10s            %2.2f%% (%d)\n",
				xdr_nfs_name(i),
				100.0 * ctxt->calls[i] / total,
				ctxt->calls[i]);
	}
	printf("\n"
	       "\tcalls total           %ld\n"
	       "\tcalls/sec             %-8.2f\n",
	       total,
	       (double) total * HZ / stats.total_time);

	xfree(ctxt);
}

