/*
 * nfs.c	NFS XDR functions.
 *
 * Copyright (C) 1996, Olaf Kirch <okir@monad.swb.de>
 */

#include <sys/types.h>
#include <string.h>
#include "nfsbench.h"
#include "xdr/nfs.h"

#define ENC	1
#define DEC	0

int	nfsstat;

/*
 * XDR encode/decode functions.
 */
static u32 *
xdr_nfs_fh(u32 *msg, int enc, nfs_fh *fh)
{
	if (enc)
		memcpy(msg, fh, sizeof(*fh));
	else
		memcpy(fh, msg, sizeof(*fh));
#if 0
	if (enc) {
		int	i;

		printf("FH   ");
		for (i = 0; i < 32; i++)
			printf("%02x", (unsigned char) fh->data[i]);
		printf("\n");
	}
#endif
	msg += (sizeof(*fh) + 3) >> 2;
	return msg;
}

static u32 *
xdr_nfs_sattr(u32 *msg, int enc, nfs_sattr *sattr)
{
	*msg++ = htonl(sattr->mode);
	*msg++ = htonl(sattr->uid);
	*msg++ = htonl(sattr->gid);
	*msg++ = htonl(sattr->size);
	*msg++ = htonl(sattr->atime.tv_sec);
	*msg++ = htonl(sattr->atime.tv_usec);
	*msg++ = htonl(sattr->mtime.tv_sec);
	*msg++ = htonl(sattr->mtime.tv_usec);
	return msg;
}

static u32 *
xdr_nfs_fname(u32 *msg, int enc, char *fname)
{
	int	len;

	if (enc) {
		len = strlen(fname);
		*msg++ = htonl(len);
		strcpy((char *) msg, fname);
	} else {
		len = ntohl(*msg++);
		memcpy(fname, msg, len);
		fname[len] = 0;
	}
	msg += ((len + 3) >> 2);
	return msg;
}

static u32 *
xdr_nfs_fattr(u32 *msg, int enc, nfs_fattr *fattr)
{
	fattr->type = ntohl(*msg++);
	fattr->mode = ntohl(*msg++);
	fattr->nlink = ntohl(*msg++);
	fattr->uid = ntohl(*msg++);
	fattr->gid = ntohl(*msg++);
	fattr->size = ntohl(*msg++);
	fattr->blocksize = ntohl(*msg++);
	fattr->rdev = ntohl(*msg++);
	fattr->blocks = ntohl(*msg++);
	fattr->fsid = ntohl(*msg++);
	fattr->fileid = ntohl(*msg++);
	fattr->atime.tv_sec = ntohl(*msg++);
	fattr->atime.tv_usec = ntohl(*msg++);
	fattr->mtime.tv_sec = ntohl(*msg++);
	fattr->mtime.tv_usec = ntohl(*msg++);
	fattr->ctime.tv_sec = ntohl(*msg++);
	fattr->ctime.tv_usec = ntohl(*msg++);
	return msg;
}

static u32 *
xdr_nfs_fstat(u32 *msg, int enc, nfs_fstat *fstat)
{
	fstat->tsize = ntohl(*msg++);
	fstat->bsize = ntohl(*msg++);
	fstat->blocks = ntohl(*msg++);
	fstat->bfree = ntohl(*msg++);
	fstat->bavail = ntohl(*msg++);
	return msg;
}

static u32
xdr_nfs_reply(u32 **msgp)
{
	rpcstat = rpc_reply(msgp);
	nfsstat = ntohl(*((*msgp)++));
	return rpcstat;
}

/*
 * NFS async procedures.
 */
u32 *
xdr_nfs_null(u32 *msg)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_NULL);
	return msg;
}

u32 *
xdr_nfs_null_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_getattr(u32 *msg, nfs_fh *fh)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_GETATTR);
	msg = xdr_nfs_fh(msg, ENC, fh);
	return msg;
}

u32 *
xdr_nfs_getattr_res(u32 *msg, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_setattr(u32 *msg, nfs_fh *fh, nfs_sattr *sattr)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_SETATTR);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_sattr(msg, ENC, sattr);
	return msg;
}

u32 *
xdr_nfs_setattr_res(u32 *msg, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_root(u32 *msg)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_ROOT);
	return msg;
}

u32 *
xdr_nfs_root_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_lookup(u32 *msg, nfs_fh *fh, char *fname)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_LOOKUP);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_fname(msg, ENC, fname);
	return msg;
}

u32 *
xdr_nfs_lookup_res(u32 *msg, nfs_fh *fh, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fh(msg, DEC, fh);
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_readlink(u32 *msg, nfs_fh *fh)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_READLINK);
	msg = xdr_nfs_fh(msg, ENC, fh);
	return msg;
}

u32 *
xdr_nfs_readlink_res(u32 *msg, char *fname)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fname(msg, DEC, fname);
	return msg;
}

u32 *
xdr_nfs_read(u32 *msg, nfs_fh *fh, u32 offset, u32 count)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_READ);
	msg = xdr_nfs_fh(msg, ENC, fh);
	*msg++ = htonl(offset);
	*msg++ = htonl(count);
	*msg++ = htonl(count);
	return msg;
}

u32 *
xdr_nfs_read_res(u32 *msg, nfs_fattr *fattr, u32 *lenp)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	*lenp = ntohl(*msg++);
	return msg;
}

u32 *
xdr_nfs_writecache(u32 *msg)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_WRITECACHE);
	return msg;
}

u32 *
xdr_nfs_writecache_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_write(u32 *msg, nfs_fh *fh, u32 offset, u32 count)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_WRITE);
	msg = xdr_nfs_fh(msg, ENC, fh);
	*msg++ = htonl(offset);
	*msg++ = htonl(offset);
	*msg++ = htonl(count);
	*msg++ = htonl(count);	/* opqaue data count */
	return msg;
}

u32 *
xdr_nfs_write_res(u32 *msg, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_create(u32 *msg, nfs_fh *fh, char *fname, nfs_sattr *sattr)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_CREATE);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_fname(msg, ENC, fname);
	msg = xdr_nfs_sattr(msg, ENC, sattr);
	return msg;
}

u32 *
xdr_nfs_create_res(u32 *msg, nfs_fh *fh, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fh(msg, DEC, fh);
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_remove(u32 *msg, nfs_fh *fh, char *fname)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_REMOVE);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_fname(msg, ENC, fname);
	return msg;
}

u32 *
xdr_nfs_remove_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_rename(u32 *msg, nfs_fh *fhfrm, char *fnfrm, nfs_fh *fhto, char *fnto)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_RENAME);
	msg = xdr_nfs_fh(msg, ENC, fhfrm);
	msg = xdr_nfs_fname(msg, ENC, fnfrm);
	msg = xdr_nfs_fh(msg, ENC, fhto);
	msg = xdr_nfs_fname(msg, ENC, fnto);
	return msg;
}

u32 *
xdr_nfs_rename_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_link(u32 *msg, nfs_fh *fhfrm, nfs_fh *fhto, char *fnto)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_LINK);
	msg = xdr_nfs_fh(msg, ENC, fhfrm);
	msg = xdr_nfs_fh(msg, ENC, fhto);
	msg = xdr_nfs_fname(msg, ENC, fnto);
	return msg;
}

u32 *
xdr_nfs_link_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

/* man, this is boring */
u32 *
xdr_nfs_symlink(u32 *msg, nfs_fh *fhfrm, char *fnfrm, char *path, nfs_sattr *sattr)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_SYMLINK);
	msg = xdr_nfs_fh(msg, ENC, fhfrm);
	msg = xdr_nfs_fname(msg, ENC, fnfrm);
	msg = xdr_nfs_fname(msg, ENC, path);
	msg = xdr_nfs_sattr(msg, ENC, sattr);
	return msg;
}

u32 *
xdr_nfs_symlink_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_mkdir(u32 *msg, nfs_fh *fh, char *fname, nfs_sattr *sattr)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_MKDIR);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_fname(msg, ENC, fname);
	msg = xdr_nfs_sattr(msg, ENC, sattr);
	return msg;
}

u32 *
xdr_nfs_mkdir_res(u32 *msg, nfs_fh *fh, nfs_fattr *fattr)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;
	msg = xdr_nfs_fh(msg, DEC, fh);
	msg = xdr_nfs_fattr(msg, DEC, fattr);
	return msg;
}

u32 *
xdr_nfs_rmdir(u32 *msg, nfs_fh *fh, char *fname)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_RMDIR);
	msg = xdr_nfs_fh(msg, ENC, fh);
	msg = xdr_nfs_fname(msg, ENC, fname);
	return msg;
}

u32 *
xdr_nfs_rmdir_res(u32 *msg)
{
	xdr_nfs_reply(&msg);
	return msg;
}

u32 *
xdr_nfs_readdir(u32 *msg, nfs_fh *fh, u32 cookie, u32 count)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_READDIR);
	msg = xdr_nfs_fh(msg, ENC, fh);
	*msg++ = htonl(cookie);
	*msg++ = htonl(count);
	return msg;
}

u32 *
xdr_nfs_readdir_res(u32 *msg, nfs_dirent_fn_t iter, void *cd)
{
	char	*name;
	u32	cookie;
	int	len;

	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;

	while (*msg++) {
		msg++;			/* fileid */
		len = ntohl(*msg++);	/* name length */
		name = (char *) msg;
		msg += (len + 3) >> 2;
		cookie = ntohl(*msg++);	/* readdir cookie */
		name[len] = '\0';
		iter(cd, name, cookie);
	}
	if (ntohl(*msg++))
		iter(cd, NULL, 0);

	return msg;
}

u32 *
xdr_nfs_statfs(u32 *msg, nfs_fh *fh)
{
	msg = rpc_header(msg, NFS_PROGRAM, NFS_VERSION, NFSPROC_STATFS);
	msg = xdr_nfs_fh(msg, ENC, fh);
	return msg;
}

u32 *
xdr_nfs_statfs_res(u32 *msg, nfs_fstat *fstat)
{
	if (xdr_nfs_reply(&msg) != RPC_SUCCESS)
		return msg;

	msg = xdr_nfs_fstat(msg, DEC, fstat);
	return msg;
}

char *
xdr_nfs_name(u32 op)
{
	switch (op) {
	case NFSPROC_NULL:	return "null";
	case NFSPROC_GETATTR:	return "getattr";
	case NFSPROC_SETATTR:	return "setattr";
	case NFSPROC_ROOT:	return "root";
	case NFSPROC_LOOKUP:	return "lookup";
	case NFSPROC_READLINK:	return "readlink";
	case NFSPROC_READ:	return "read";
	case NFSPROC_WRITECACHE:return "writecache";
	case NFSPROC_WRITE:	return "write";
	case NFSPROC_CREATE:	return "create";
	case NFSPROC_REMOVE:	return "remove";
	case NFSPROC_RENAME:	return "rename";
	case NFSPROC_LINK:	return "link";
	case NFSPROC_SYMLINK:	return "symlink";
	case NFSPROC_MKDIR:	return "mkdir";
	case NFSPROC_RMDIR:	return "rmdir";
	case NFSPROC_READDIR:	return "readdir";
	case NFSPROC_STATFS:	return "statfs";
	}
	return "unknown";
}
